#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Base/Parent Wrapper class.
"""

from abc import ABC


class TimeSHAPWrapper(ABC):
    """
    Base class for TimeSHAP model wrappers.
    """
    def __init__(self, model, batch_budget):
        self.model = model
        self.batch_budget = batch_budget

    def __call__(self, *args, **kwargs):
        return self.model(*args, *kwargs)
